@@ -7,75 +7,42 @@ and functions to compute normalized mutual information (NMI) and relevance redun
 
 """
 
+from __future__ import annotations
 import os
 import logging
 
+from pathlib import Path
+from collections import namedtuple
+from typing import Dict, List, Union, Optional, Callable, Hashable, Iterable, Tuple
+from functools import partial
+
 from pymatgen import Structure
-from pymatgen.core.periodic_table import Element
-from sklearn.feature_selection import mutual_info_regression
+from sklearn.feature_selection import mutual_info_regression, mutual_info_classif
 import pandas as pd
 import numpy as np
+import tqdm
 
-from matminer.featurizers.base import MultipleFeaturizer
-from matminer.featurizers.composition import (
-    AtomicOrbitals,
-    AtomicPackingEfficiency,
-    BandCenter,
-    # CohesiveEnergy, - This descriptor was not used in the paper preset
-    # ElectronAffinity, - This descriptor was not used in the paper preset
-    ElectronegativityDiff,
-    ElementFraction,
-    ElementProperty,
-    IonProperty,
-    Miedema,
-    OxidationStates,
-    Stoichiometry,
-    TMetalFraction,
-    ValenceOrbital,
-    YangSolidSolution,
-)
-
-from matminer.featurizers.conversions import CompositionToOxidComposition
-from matminer.featurizers.structure import (
-    BagofBonds,
-    BondFractions,
-    ChemicalOrdering,
-    CoulombMatrix,
-    DensityFeatures,
-    EwaldEnergy,
-    GlobalSymmetryFeatures,
-    MaximumPackingEfficiency,
-    PartialRadialDistributionFunction,
-    RadialDistributionFunction,
-    SineCoulombMatrix,
-    SiteStatsFingerprint,
-    StructuralHeterogeneity,
-    XRDPowderPattern,
-)
-from matminer.featurizers.site import (
-    AGNIFingerprints,
-    AverageBondAngle,
-    AverageBondLength,
-    BondOrientationalParameter,
-    ChemEnvSiteFingerprint,
-    CoordinationNumber,
-    CrystalNNFingerprint,
-    GaussianSymmFunc,
-    GeneralizedRadialDistributionFunction,
-    LocalPropertyDifference,
-    OPSiteFingerprint,
-    VoronoiFingerprint,
-)
-from pymatgen.analysis.local_env import VoronoiNN
-from typing import Dict, List, Union, Optional, Callable, Hashable, Iterable
+from modnet.featurizers import MODFeaturizer
+from modnet import __version__
 
 DATABASE = pd.DataFrame([])
 
+Dataset = namedtuple("Dataset", ("url", "filename", "md5"))
+DATASETS = {
+    "MP_2018.6": Dataset(
+        url="https://ndownloader.figshare.com/files/24364571",
+        filename="MP_2018.6.zip",
+        md5="06280c4e539508bbcc5266f07698f8d1"
+    ),
+}
+
+EPS = 1e-16
+
 logging.getLogger().setLevel(logging.INFO)
 
 
 def nmi_target(df_feat: pd.DataFrame, df_target: pd.DataFrame,
-               drop_constant_features: bool = True, **kwargs) -> pd.DataFrame:
+               task_type: str = "regression", drop_constant_features: bool = True, **kwargs) -> pd.DataFrame:
     """
     Computes the Normalized Mutual Information (NMI) between a list of
     input features and a target variable.
@@ -86,6 +53,7 @@ def nmi_target(df_feat: pd.DataFrame, df_target: pd.DataFrame,
         df_target (pandas.DataFrame): Dataframe containing the target variable.
             This DataFrame should contain only one column and have the same
             size as `df_feat`.
+        task_type (integer): 0 for regression, 1 for classification
         drop_constant_features (bool): If True, the features that are constant
             across the entire data set will be dropped.
         **kwargs: Keyword arguments to be passed down to the
@@ -113,23 +81,32 @@ def nmi_target(df_feat: pd.DataFrame, df_target: pd.DataFrame,
         to_drop = frange[frange == 0].index
         df_feat = df_feat.drop(to_drop, axis=1)
 
+    # Take right MI fun depending on regression / classification
+    if task_type == "regression":
+        _mifun = mutual_info_regression
+        _self_mifun = mutual_info_regression
+    elif task_type == "classification":
+        _mifun = mutual_info_classif
+        _self_mifun = partial(mutual_info_classif, discrete_features=True)
+
     # Prepare the output DataFrame and compute the mutual information
     target_name = df_target.columns[0]
-    out_df = pd.DataFrame([], columns=[target_name], index=df_feat.columns)
-    out_df.loc[:, target_name] = (mutual_info_regression(df_feat, df_target[target_name], **kwargs))
+    mutual_info = pd.DataFrame([], columns=[target_name], index=df_feat.columns)
+
+    mutual_info.loc[:, target_name] =_mifun(df_feat, df_target[target_name], **kwargs)
 
     # Compute the "self" mutual information (i.e. information entropy) of the target variable and of the input features
-    target_mi = mutual_info_regression(df_target[target_name].values.reshape(-1, 1),
+    target_mi = _self_mifun(df_target[target_name].values.reshape(-1, 1),
                                        df_target[target_name], **kwargs)[0]
     diag = {}
     for x in df_feat.columns:
         diag[x] = (mutual_info_regression(df_feat[x].values.reshape(-1, 1), df_feat[x], **kwargs))[0]
 
     # Normalize the mutual information
-    for x in out_df.index:
-        out_df.loc[x, target_name] = out_df.loc[x, target_name] / ((target_mi + diag[x])/2)
+    for x in mutual_info.index:
+        mutual_info.loc[x, target_name] = mutual_info.loc[x, target_name] / ((target_mi + diag[x])/2)
 
-    return out_df
+    return mutual_info
 
 
 def get_cross_nmi(df_feat: pd.DataFrame, **kwargs) -> pd.DataFrame:
@@ -145,31 +122,45 @@ def get_cross_nmi(df_feat: pd.DataFrame, **kwargs) -> pd.DataFrame:
 
     Returns:
         pd.DataFrame: pandas.DataFrame containing the Normalized Mutual Information between features.
+
     """
+
+    if kwargs.get("random_state"):
+        seed = kwargs.pop("random_state")
+    else:
+        seed = np.random.RandomState()
+
     # Prepare the output DataFrame and compute the mutual information
-    out_df = pd.DataFrame([], columns=df_feat.columns, index=df_feat.columns)
-    for ifeat, feat_name in enumerate(out_df.columns, start=1):
-        logging.info('Computing MI of feature #{:d}/{:d} ({}) with all other features'.format(ifeat,
-                                                                                              len(out_df.columns),
-                                                                                              feat_name))
-        out_df.loc[:, feat_name] = (mutual_info_regression(df_feat, df_feat[feat_name], **kwargs))
+    mutual_info = pd.DataFrame([], columns=df_feat.columns, index=df_feat.columns)
 
     # Compute the "self" mutual information (i.e. information entropy) of the features
     logging.info('Computing "self" MI (i.e. information entropy) of features')
     diag = {}
-    for x in df_feat.columns:
-        diag[x] = (mutual_info_regression(df_feat[x].values.reshape(-1, 1), df_feat[x], **kwargs))[0]
+    for x_feat in df_feat.columns:
+        diag[x_feat] = mutual_info_regression(
+            df_feat[x_feat].values.reshape(-1, 1),
+            df_feat[x_feat],
+            random_state=seed,
+            **kwargs
+        )[0]
+        if diag[x_feat] < 0.2 or abs(df_feat[x_feat].max() - df_feat[x_feat].min()) < EPS:
+            mutual_info.loc[x_feat, x_feat] = np.nan
+            diag[x_feat] = np.nan
+        else:
+            mutual_info.loc[x_feat, x_feat] = 1.0
 
-    # Normalize the mutual information between features
-    logging.info('Normalizing MI')
-    for feat1 in out_df.index:
-        for feat2 in out_df.columns:
-            out_df.loc[feat1, feat2] = out_df.loc[feat1, feat2] / ((diag[feat1] + diag[feat2]) / 2)
-        logging.debug('  => Computed NMI of feature "{}" with all other features :\n'
-                      '{}'.format(feat1, '\n'.join(['      {} : {:.4f}'.format(feat2,
-                                                                               out_df.loc[feat1, feat2])
-                                                    for feat2 in out_df.loc[feat1, :].index])))
-    return out_df
+    logging.info('Computing cross NMI between all features...')
+    for idx, x_feat in tqdm.tqdm(enumerate(mutual_info.columns), total=len(mutual_info.columns)):
+        for _, y_feat in enumerate(mutual_info.columns[idx+1:]):
+            I_xy = mutual_info_regression(
+                df_feat[y_feat].values.reshape(-1, 1),
+                df_feat[x_feat],
+                random_state=seed,
+                **kwargs
+            )[0] / (0.5 * (diag[x_feat] + diag[y_feat]))
+            mutual_info.loc[y_feat, x_feat] = mutual_info.loc[x_feat, y_feat] = I_xy
+
+    return mutual_info
 
 
 def get_rr_p_parameter_default(nn: int) -> float:
@@ -336,19 +327,22 @@ def get_features_relevance_redundancy(
 
 def get_features_dyn(n_feat, cross_nmi, target_nmi):
 
-    first_feature = target_nmi.nlargest(1).index[0]
-    feature_set = [first_feature]
-
-    get_p = get_rr_p_parameter_default
-    get_c = get_rr_c_parameter_default
-
     missing = [x for x in cross_nmi.index if x not in target_nmi.index]
     cross_nmi = cross_nmi.drop(missing, axis=0).drop(missing, axis=1)
 
+    missing = [x for x in target_nmi.index if x not in cross_nmi.index]
+    target_nmi = target_nmi.drop(missing, axis=0)
+    target_nmi = target_nmi.replace([np.inf, -np.inf, np.nan], 0)
+
+    first_feature = target_nmi.nlargest(1).index[0]
+    feature_set = [first_feature]
+    get_p = get_rr_p_parameter_default
+    get_c = get_rr_c_parameter_default
+
     if n_feat == -1:
         n_feat = len(cross_nmi.index)
     else:
-        n_feat = min(len(cross_nmi.index),n_feat)
+        n_feat = min(len(cross_nmi.index), n_feat)
 
     for n in range(n_feat-1):
         if (n+1) % 50 == 0:
@@ -358,7 +352,7 @@ def get_features_dyn(n_feat, cross_nmi, target_nmi):
         c = get_c(n)
 
         score = cross_nmi.copy()
-        #score = score.loc[target_mi.index, target_mi.index]
+        # score = score.loc[target_mi.index, target_mi.index]
         score = score.drop(feature_set, axis=0)
         score = score[feature_set]
 
@@ -403,218 +397,6 @@ def merge_ranked(lists: List[List[Hashable]]) -> List[Hashable]:
     return ranked_list
 
 
-def clean_df(df):
-    """ Cleans dataframe by dropping missing values, replacing NaN's and infinities
-    and selecting only columns containing numerical data.
-
-    Args:
-        df (pd.DataFrame): the dataframe to clean.
-
-    Returns:
-        pd.DataFrame: the cleaned dataframe.
-
-    """
-
-    df = df.dropna(axis=1, how='all')
-    df = df.replace([np.inf, -np.inf, np.nan], -1)
-    df = df.select_dtypes(include='number')
-
-    return df
-
-
-def featurize_composition(df: pd.DataFrame) -> pd.DataFrame:
-    """ Decorate input `pandas.DataFrame` of structures with composition
-    features from matminer.
-
-    Currently applies the set of all matminer composition features.
-
-    Args:
-        df (pandas.DataFrame): the input dataframe with `"structure"`
-            column containing `pymatgen.Structure` objects.
-
-    Returns:
-        pandas.DataFrame: the decorated DataFrame.
-
-    """
-    logging.info("Applying composition featurizers...")
-    df = df.copy()
-    df['composition'] = df['structure'].apply(lambda s: s.composition)
-    featurizer = MultipleFeaturizer([ElementProperty.from_preset("magpie"),
-                                     AtomicOrbitals(),
-                                     BandCenter(),
-                                     # ElectronAffinity(), - This descriptor was not used in the paper preset
-                                     Stoichiometry(),
-                                     ValenceOrbital(),
-                                     IonProperty(),
-                                     ElementFraction(),
-                                     TMetalFraction(),
-                                     # CohesiveEnergy(), - This descriptor was not used in the paper preset
-                                     Miedema(),
-                                     YangSolidSolution(),
-                                     AtomicPackingEfficiency(),
-                                     ])
-
-    df = featurizer.featurize_dataframe(df, "composition", multiindex=True, ignore_errors=True)
-    df.columns = df.columns.map('|'.join).str.strip('|')
-
-    ox_featurizer = MultipleFeaturizer([OxidationStates(),
-                                        ElectronegativityDiff()
-                                        ])
-
-    df = CompositionToOxidComposition().featurize_dataframe(df, "Input Data|composition")
-
-    df = ox_featurizer.featurize_dataframe(df, "composition_oxid", multiindex=True, ignore_errors=True)
-    df = df.rename(columns={'Input Data': ''})
-    df.columns = df.columns.map('|'.join).str.strip('|')
-
-    _orbitals = {"s": 1, "p": 2, "d": 3, "f": 4}
-
-    df['AtomicOrbitals|HOMO_character'] = df['AtomicOrbitals|HOMO_character'].map(_orbitals)
-    df['AtomicOrbitals|LUMO_character'] = df['AtomicOrbitals|LUMO_character'].map(_orbitals)
-
-    df['AtomicOrbitals|HOMO_element'] = df['AtomicOrbitals|HOMO_element'].apply(
-        lambda x: -1 if not isinstance(x, str) else Element(x).Z
-    )
-    df['AtomicOrbitals|LUMO_element'] = df['AtomicOrbitals|LUMO_element'].apply(
-        lambda x: -1 if not isinstance(x, str) else Element(x).Z
-    )
-
-    df = df.replace([np.inf, -np.inf, np.nan], 0)
-
-    return clean_df(df)
-
-
-def featurize_structure(df: pd.DataFrame) -> pd.DataFrame:
-    """ Decorate input `pandas.DataFrame` of structures with structural
-    features from matminer.
-
-    Currently applies the set of all matminer structure features.
-
-    Args:
-        df (pandas.DataFrame): the input dataframe with `"structure"`
-            column containing `pymatgen.Structure` objects.
-
-    Returns:
-        pandas.DataFrame: the decorated DataFrame.
-
-    """
-
-    logging.info("Applying structure featurizers...")
-
-    df = df.copy()
-
-    structure_features = [
-         DensityFeatures(),
-         GlobalSymmetryFeatures(),
-         RadialDistributionFunction(),
-         CoulombMatrix(),
-         PartialRadialDistributionFunction(),
-         SineCoulombMatrix(),
-         EwaldEnergy(),
-         BondFractions(),
-         StructuralHeterogeneity(),
-         MaximumPackingEfficiency(),
-         ChemicalOrdering(),
-         XRDPowderPattern(),
-         BagofBonds()
-    ]
-
-    featurizer = MultipleFeaturizer([feature.fit(df["structure"]) for feature in structure_features])
-
-    df = featurizer.featurize_dataframe(df, "structure", multiindex=True, ignore_errors=True)
-    df.columns = df.columns.map('|'.join).str.strip('|')
-
-    dist = df["RadialDistributionFunction|radial distribution function"][0]['distances'][:50]
-    for i, d in enumerate(dist):
-        _rdf_key = "RadialDistributionFunction|radial distribution function|d_{:.2f}".format(d)
-        df[_rdf_key] = df["RadialDistributionFunction|radial distribution function"].apply(lambda x: x['distribution'][i])
-
-    df = df.drop("RadialDistributionFunction|radial distribution function", axis=1)
-
-    _crystal_system = {
-        "cubic": 1, "tetragonal": 2, "orthorombic": 3,
-        "hexagonal": 4, "trigonal": 5, "monoclinic": 6, "triclinic": 7
-    }
-
-    df["GlobalSymmetryFeatures|crystal_system"] = df["GlobalSymmetryFeatures|crystal_system"].map(_crystal_system)
-    df["GlobalSymmetryFeatures|is_centrosymmetric"] = df["GlobalSymmetryFeatures|is_centrosymmetric"].map(int)
-
-    return clean_df(df)
-
-
-def featurize_site(df: pd.DataFrame, site_stats=("mean", "std_dev")) -> pd.DataFrame:
-    """ Decorate input `pandas.DataFrame` of structures with site
-    features from matminer.
-
-    Currently creates the set of all matminer structure features with
-    the `matminer.featurizers.structure.SiteStatsFingerprint`.
-
-    Args:
-        df (pandas.DataFrame): the input dataframe with `"structure"`
-            column containing `pymatgen.Structure` objects.
-        site_stats (Tuple[str]): the matminer site stats to use in the
-            `SiteStatsFingerprint` for all features.
-
-    Returns:
-        pandas.DataFrame: the decorated DataFrame.
-
-    """
-
-    logging.info("Applying site featurizers...")
-
-    df = df.copy()
-    df.columns = ["Input data|" + x for x in df.columns]
-
-    site_fingerprints = (
-        AGNIFingerprints(),
-        GeneralizedRadialDistributionFunction.from_preset("gaussian"),
-        OPSiteFingerprint(),
-        CrystalNNFingerprint.from_preset("ops"),
-        VoronoiFingerprint(),
-        GaussianSymmFunc(),
-        ChemEnvSiteFingerprint.from_preset("simple"),
-        CoordinationNumber(),
-        LocalPropertyDifference(),
-        BondOrientationalParameter(),
-        AverageBondLength(VoronoiNN()),
-        AverageBondAngle(VoronoiNN())
-    )
-
-    for fingerprint in site_fingerprints:
-        site_stats_fingerprint = SiteStatsFingerprint(
-            fingerprint,
-            stats=site_stats
-        )
-
-        df = site_stats_fingerprint.featurize_dataframe(
-            df,
-            "Input data|structure",
-            multiindex=False,
-            ignore_errors=True
-        )
-
-        fingerprint_name = fingerprint.__class__.__name__
-
-        # rename some features for backwards compatibility with pretrained models
-        if fingerprint_name == "GeneralizedRadialDistributionFunction":
-            fingerprint_name = "GeneralizedRDF"
-        elif fingerprint_name == "AGNIFingerprints":
-            fingerprint_name = "AGNIFingerPrint"
-        elif fingerprint_name == "BondOrientationalParameter":
-            fingerprint_name = "BondOrientationParameter"
-        elif fingerprint_name == "GaussianSymmFunc":
-            fingerprint_name = "ChemEnvSiteFingerprint|GaussianSymmFunc"
-
-        if "|" not in fingerprint_name:
-            fingerprint_name += "|"
-
-        df.columns = [f"{fingerprint_name}{x}" if "|" not in x else x for x in df.columns]
-
-    df = df.loc[:, (df != 0).any(axis=0)]
-
-    return clean_df(df)
-
-
 class MODData:
     """ The MODData class takes takes a list of `pymatgen.Structure`
     objects and creates a `pandas.DataFrame` that contains many matminer
@@ -631,10 +413,17 @@ class MODData:
             computed features per structure, indexed by ID.
         optimal_features (List[str]): if feature selection has been performed
             this attribute stores a list of the selected features.
-        optimal_features_by_target (Dict[str, List[str]]): if feature selection has been performed
-            this attribute stores a list of the selected features, broken down by
-            target property.
-
+        optimal_features_by_target (Dict[str, List[str]]): If feature selection has been performed
+            this attribute stores a list of the selected features, broken down by target property.
+        featurizer (MODFeaturizer): the class used to featurize the data.
+        __modnet_version__ (str): The MODNet version number used to create the object
+        cross_nmi (pd.DataFrame): If feature selection has been performed, this attribute
+            stores the normalized mutual information between all features.
+        num_classes (pd.DataFrame): If feature selection has been performed, this attribute
+            stores the normalized mutual information between all features and all targets.
+        num_classes: Dictionary defining the target types (classification or regression).
+                Should be constructed as follows: key: string giving the target name; value: integer n,
+                 with n=0 for regression and n>=2 for classification with n the number of classes.
     """
 
     def __init__(
@@ -643,7 +432,9 @@ class MODData:
         targets: Optional[Union[List[float], np.ndarray]] = None,
         target_names: Optional[Iterable] = None,
         structure_ids: Optional[Iterable] = None,
+        num_classes: Optional[Dict[str,int]] = None,
         df_featurized: Optional[pd.DataFrame] = None,
+        featurizer: Optional[Union[MODFeaturizer, str]] = None,
     ):
         """ Initialise the MODData object either from a list of structures
         or from an already featurized dataframe. Prediction targets per
@@ -657,33 +448,53 @@ class MODData:
              is a ndarray where each column corresponds to a target, i.e. of shape (n_materials,n_targets).
             target_names: optional Iterable (e.g. list) of names of target properties to use in the dataframe.
             structure_ids: optional Iterable of unique IDs to use instead of generated integers.
+            num_classes: Dictionary defining the target types (classification or regression).
+                Should be constructed as follows: key: string giving the target name; value: integer n,
+                 with n=0 for regression and n>=2 for classification with n the number of classes.
             df_featurized: optional featurized dataframe to use instead of
                 featurizing a new one. Should be passed without structures.
+            featurizer: optional MODFeaturizer object to use for featurization, or string
+                preset to look up in presets dictionary.
 
         """
 
+        from modnet.featurizers.presets import FEATURIZER_PRESETS
+        self.__modnet_version__ = __version__
         self.df_featurized = df_featurized
+        self.cross_nmi = None
 
         if structures is not None and self.df_featurized is not None:
-            raise RuntimeError(
-                "Only one of `structures` or `df_featurized` should be passed to `MODData`."
-            )
+            if len(structures) != len(self.df_featurized):
+                raise RuntimeError("Mismatched shape of structures and passed df_featurized")
+
         if structures is None and self.df_featurized is None:
             raise RuntimeError(
                 "At least one of `structures` or `df_featurized` should be passed to `MODData`."
             )
 
         if targets is not None:
-            targets = np.array(targets).reshape((len(targets),-1))
+            targets = np.array(targets).reshape((len(targets), -1))
 
         if structures is not None and targets is not None:
             if np.shape(targets)[0] != len(structures):
                 raise ValueError(f"Targets must have same length as structures: {np.shape(targets)} vs {len(structures)}")
 
+        if isinstance(featurizer, str):
+            self.featurizer = FEATURIZER_PRESETS.get(featurizer)()
+            if self.featurizer is None:
+                raise RuntimeError("Requested preset {featurizer} not found in available presets: {FEATURIZER_PRESETS.keys()}")
+        elif isinstance(featurizer, MODFeaturizer):
+            self.featurizer = featurizer
+        elif featurizer is None and self.df_featurized is None:
+            self.featurizer = FEATURIZER_PRESETS["DeBreuck2020"]()
+
+        if self.featurizer is not None:
+            logging.info(f"Loaded {self.featurizer.__class__.__name__} featurizer.")
+
         if target_names is not None:
             if np.shape(targets)[-1] != len(target_names):
                 raise ValueError("Target names must be supplied for every target.")
-        else:
+        elif targets is not None:
             target_names = ['prop'+str(i) for i in range(len(targets))]
 
         if structure_ids is not None:
@@ -697,17 +508,23 @@ class MODData:
                 raise ValueError("List of IDs (`structure_ids`) must have same length as list of structure.")
 
         else:
-            structure_ids = [f"id{i}" for i in range(len(structures))]
+            num_entries = len(structures) if structures is not None else len(df_featurized)
+            structure_ids = [f"id{i}" for i in range(num_entries)]
 
         if targets is not None:
             # set up dataframe for targets with columns (id, property_1, ..., property_n)
-            self.df_targets = pd.DataFrame(targets,index=structure_ids,columns=target_names)
+            self.df_targets = pd.DataFrame(targets, index=structure_ids, columns=target_names)
 
         # set up dataframe for structures with columns (id, structure)
         self.df_structure = pd.DataFrame({'id': structure_ids, 'structure': structures})
         self.df_structure.set_index('id', inplace=True)
 
-    def featurize(self, fast: bool = False, db_file: str = 'feature_database.pkl'):
+        self.num_classes = {name: 0 for name in self.target_names}
+        if num_classes is not None:
+            for k,v in num_classes.items():
+                self.num_classes[k] = v
+
+    def featurize(self, fast: bool = False, db_file: str = 'feature_database.pkl', n_jobs=None):
         """ For the input structures, construct many matminer features
         and save a featurized dataframe. If `db_file` is specified, this
         method will try to load previous feature calculations for each
@@ -727,6 +544,9 @@ class MODData:
         df_done = None
         df_todo = None
 
+        if n_jobs is not None:
+            self.featurizer.set_n_jobs(n_jobs)
+
         if self.df_featurized is not None:
             raise RuntimeError("Not overwriting existing featurized dataframe.")
 
@@ -747,21 +567,17 @@ class MODData:
         if fast and not df_done.empty:
             # if any are left to compute, do them
             if len(df_todo) > 0:
-                df_composition = featurize_composition(df_todo)
-                df_structure = featurize_structure(df_todo)
-                df_site = featurize_site(df_todo)
-                df_final = df_done.append(df_composition.join(df_structure.join(df_site, lsuffix="l"), rsuffix="r"))
+                df_finished = self.featurizer.featurize(df_todo)
+                df_final = df_done.append(df_finished)
                 df_final = df_final.reindex(self.structure_ids)
+
             # otherwise, all structures were successfully loaded
             else:
                 df_final = df_done
 
         # otherwise, no structures were loaded, so we need to compute all
         else:
-            df_composition = featurize_composition(self.df_structure)
-            df_structure = featurize_structure(self.df_structure)
-            df_site = featurize_site(self.df_structure)
-            df_final = df_composition.join(df_structure.join(df_site, lsuffix="l"), rsuffix="r")
+            df_final = self.featurizer.featurize(self.df_structure)
 
         df_final = df_final.replace([np.inf, -np.inf, np.nan], 0)
 
@@ -790,27 +606,36 @@ class MODData:
         ranked_lists = []
         optimal_features_by_target = {}
 
+        if cross_nmi is not None:
+            self.cross_nmi = cross_nmi
+        elif getattr(self, "cross_nmi") is None:
+            self.cross_nmi = None
+
         # Loading mutual information between features
-        if cross_nmi is None:
+        if self.cross_nmi is None:
             this_dir, this_filename = os.path.split(__file__)
             dp = os.path.join(this_dir, "data", "Features_cross")
             if os.path.isfile(dp):
-                cross_nmi = pd.read_pickle(dp)
+                logging.info("Loading cross NMI from 'Features_cross' file.")
+                self.cross_nmi = pd.read_pickle(dp)
 
-        #if cross_nmi is None:
-        #        logging.info('Computing cross NMI between all features...')
-        #        df = self.df_featurized.copy()
-        #        cross_nmi = get_cross_nmi(df)
+        if self.cross_nmi is None:
+            df = self.df_featurized.copy()
+            self.cross_nmi = get_cross_nmi(df)
 
         for i, name in enumerate(self.names):
-            logging.info("Starting target {}/{}: {} ...".format(i+1, len(self.names), self.names[i]))
+            logging.info(f"Starting target {i+1}/{len(self.names)}: {self.names[i]} ...")
 
             # Computing mutual information with target
             logging.info("Computing mutual information between features and target...")
-            y_nmi = nmi_target(self.df_featurized, self.df_targets[[name]])[name]
+            if self.num_classes[name] >=2:
+                task_type = "classification"
+            else:
+                task_type = "regression"
+            self.target_nmi = nmi_target(self.df_featurized, self.df_targets[[name]], task_type)[name]
 
             logging.info('Computing optimal features...')
-            optimal_features_by_target[name] = get_features_dyn(n, cross_nmi, y_nmi)
+            optimal_features_by_target[name] = get_features_dyn(n, self.cross_nmi, self.target_nmi)
             ranked_lists.append(optimal_features_by_target[name])
 
             logging.info("Done with target {}/{}: {}.".format(i+1, len(self.targets), name))
@@ -841,12 +666,17 @@ class MODData:
         """ Returns the list of prediction target field names. """
         return list(self.df_targets)
 
+    @property
+    def target_names(self) -> List[str]:
+        """ Returns the list of prediction target field names. """
+        return list(self.df_targets)
+
     @property
     def structure_ids(self) -> List[str]:
         """ Returns the list of prediction target field names. """
         return list(self.df_structure.index)
 
-    def save(self, filename):
+    def save(self, filename: str):
         """ Pickle the contents of the `MODData` object
         so that it can be loaded in  with `MODData.load()`.
 
@@ -858,15 +688,35 @@ class MODData:
         logging.info(f'Data successfully saved as {filename}!')
 
     @staticmethod
-    def load(filename):
+    def load(filename: Union[str, Path]) -> MODData:
         """ Load `MODData` object pickled by the `.save(...)` method.
 
         If the filename ends in "tgz", "bz2" or "zip", the pickle
         will be decompressed accordingly by `pandas.read_pickle(...)`.
 
         """
-        pickled_data = pd.read_pickle(filename)
+        pickled_data = None
+
+        if isinstance(filename, Path):
+            filename = str(filename)
+
+        # handle .zip files explicitly for OS X/macOS compatibility
+        if filename.endswith(".zip"):
+            from zipfile import ZipFile
+            with ZipFile(filename, "r") as zf:
+                namelist = zf.namelist()
+                _files = [_ for _ in namelist if not _.startswith("__MACOSX/") or _.startswith(".DS_STORE")]
+                if len(_files) == 1:
+                    with zf.open(_files.pop()) as f:
+                        pickled_data = pd.read_pickle(f)
+
+        if pickled_data is None:
+            pickled_data = pd.read_pickle(filename)
+
         if isinstance(pickled_data, MODData):
+            if not hasattr(pickled_data, "__modnet_version__"):
+                pickled_data.__modnet_version__ = "<=0.1.7"
+            logging.info(f"Loaded {pickled_data} object, created with modnet version {pickled_data.__modnet_version__}")
             return pickled_data
 
         raise ValueError(
@@ -874,6 +724,47 @@ class MODData:
             f"instead found {pickled_data.__class__.__name__}."
         )
 
+    @classmethod
+    def load_precomputed(cls, dataset_name: str):
+        """ Load a `MODData` object from a pre-computed dataset.
+
+        Note:
+            Datasets may require significant (~10 GB) amounts of memory
+            to load.
+
+        Arguments:
+            dataset: the name of the precomputed dataset to load.
+                Currently available: 'MP_2018.6'.
+
+        Returns:
+            MODData: the precomputed dataset.
+
+        """
+        import urllib
+
+        if dataset_name not in DATASETS:
+            raise ValueError(f"No dataset {dataset_name} found, must be one of {list(DATASETS.keys())}")
+
+        dataset = DATASETS[dataset_name]
+
+        model_path = Path(__file__).parent.parent.joinpath(f"moddata/{dataset.filename}")
+        if not model_path.is_file():
+            logging.info(f"Downloading featurized dataset {dataset_name} from {dataset.url} into {model_path}")
+            try:
+                zip_file, response = urllib.request.urlretrieve(dataset.url, model_path)
+            except (urllib.error.URLError, urllib.error.HTTPError) as exc:
+                raise ValueError(f"There was a problem downloading {dataset.url}: {exc.reason}")
+
+        if dataset.md5 is not None:
+            from modnet.utils import get_hash_of_file
+            file_md5 = get_hash_of_file(model_path, algo="md5")
+            if file_md5 != dataset.md5:
+                raise RuntimeError(
+                    "Precomputed MODData did not match expected MD5 from {dataset.url}, will not unpickled."
+                )
+
+        return cls.load(str(model_path))
+
     def get_structure_df(self):
         return self.df_structure
 
@@ -887,4 +778,49 @@ class MODData:
         return self.optimal_features
 
     def get_optimal_df(self):
-        return self.df_featurized[self.optimal_features].join(self.targets)
+        return self.df_featurized[self.optimal_features].join(self.get_target_df())
+
+    def split(self, train_test_split: Tuple[List[int], List[int]]) -> Tuple[MODData, MODData]:
+        """ Create two new MODData's that contain only the data corresponding
+        to the indices passed in the `train_test_split` tuple.
+
+        Arguments:
+            train_test_split: A tuple containing two lists of integers: the
+                indices of the training data and test data respectively.
+
+        Returns:
+            The training MODData and the test MODData as a tuple.
+
+        """
+
+        train, test = train_test_split
+        train_moddata = self.from_indices(train)
+        test_moddata = self.from_indices(test)
+        return train_moddata, test_moddata
+
+    def from_indices(self, indices: List[int]) -> MODData:
+        """ Create a new MODData that contains only the data at the given
+        rows indices provided.
+
+        Arguments:
+            indices: The list of integers corresponding to the rows.
+
+        Returns:
+            A `MODData` containing only the rows passed.
+
+        """
+        split_data = MODData.__new__(MODData)
+        extensive_dataframes = ("df_structure", "df_targets", "df_featurized")
+        for attr in extensive_dataframes:
+            setattr(split_data, attr, getattr(self, attr).iloc[indices])
+
+        for attr in [_ for _ in dir(self) if _ not in extensive_dataframes]:
+            if not callable(getattr(self, attr)) and not attr.startswith("__"):
+                try:
+                    setattr(split_data, attr, getattr(self, attr))
+                except AttributeError:
+                    pass
+
+        split_data.__modnet_version__ = __version__
+
+        return split_data
